#pragma once

#include <cstdint>
#include <cassert>

#include <string>
#include <vector>
#include <ostream>
#include <functional>
#include <stdexcept>
#include <iterator>

#include "kronos_abi.h"

#ifdef __APPLE__
#include <alloca.h>
#endif

namespace Kronos {
	/**
	* Utility: shared reference and value semantics
	*/

	static void _CheckLastError();
	template <typename T> T _CheckResult(T&& v) { if (v == 0) _CheckLastError(); return v; }

	template <class C> class Shared;
	template <class C> class Value;

	template <class C> class Pointer {
		friend class Shared<C>;
		friend class Value<C>;
	protected:
		C* i;
		Pointer(C* ptr):i(ptr) { }
	public:
		C* operator->() { return i; }
		const C* operator->() const { return i; }
		operator const C&() const { return *i; }
		operator C&() { return *i; }
		const C& operator*() const { return *i; }
		C& operator*() { return *i; }
		C* Get() noexcept { return i; }
		const C* Get() const { return i; }
		operator bool() const { return i != nullptr; }
		bool operator==(const Pointer<C>& r) { return i == r.Get(); }
		bool operator!=(const Pointer<C>& r) { return i != r.Get( ); }
	};

	template <class C> class Shared : public Pointer<C> {
	protected:
		using Pointer<C>::i;
	public:
		Shared& operator=(Shared s) { std::swap(i, s.i); return *this; }
		Shared(const Shared& src):Pointer<C>(src.i) { if (i) i->Retain(); }
		Shared(Shared&& src):Pointer<C>(src.i) { src.i = nullptr; }
		Shared(C* src = nullptr):Pointer<C>(src) { if (i) i->Retain(); }
		~Shared() { if (i) i->Release(); }
		size_t GetHash() const { return std::hash<const C*>()(this->Get()); }
	};

	template <class C> class Value : public Pointer<C> {
	protected:
		using Pointer<C>::i;
	public:
		Value& operator=(Value<C>&& s) { std::swap(i, s.i); return *this; }
		Value& operator=(const Value<C>& s) { return operator=(Value<C>(s)); }
		Value(const Value<C>& src):Pointer<C>(static_cast<C*>(src?src->Clone():nullptr)) { _CheckLastError(); }
		Value(const C& src):Pointer<C>(static_cast<C*>(src.Clone())) { _CheckLastError(); }
		Value(Value<C>&& src):Pointer<C>(src.i) { src.i = nullptr; }
		Value(C* src = nullptr):Pointer<C>(src) { }
		~Value() { if (i) i->Delete(); }
	};


	template<typename BASE, typename STDEX> class ErrorImpl : public STDEX, public BASE {
	public:
		ErrorImpl(const char *message):STDEX(message) { }
		const char *GetErrorMessage() const noexcept { return STDEX::what(); }
		int GetErrorCode() const noexcept { return -1; }
		const char* GetSourceFilePosition() const noexcept { return ""; }
		IError::ErrorType GetErrorType() const noexcept { return BASE::GetClassErrorType();  }
		void Delete() noexcept { delete this; }
		IError* Clone() const noexcept { return new ErrorImpl(*this); }
	};

	template <typename BASE> class ProgramErrorImpl : public std::invalid_argument, public BASE {
		int code; const char* sfp;
	public:
		ProgramErrorImpl(int code, const char *what, const char* sfp):invalid_argument(what),code(code),sfp(sfp) { }
		const char *GetErrorMessage() const noexcept { return what(); }
		const char *GetSourceFilePosition() const noexcept { return sfp; }
		int GetErrorCode() const noexcept { return code; }
		IError::ErrorType GetErrorType() const noexcept { return BASE::GetClassErrorType(); }
		void Delete() noexcept { delete this; }
		IError* Clone() const noexcept { return new ProgramErrorImpl(*this); }
	};

	using InternalError = ErrorImpl<IInternalError, std::logic_error>;
	using RuntimeError = ErrorImpl<IRuntimeError, std::runtime_error>;
	using ProgramError = ProgramErrorImpl<IProgramError>;
	using SyntaxError = ProgramErrorImpl<ISyntaxError>;
	using TypeError = ProgramErrorImpl<ITypeError>;

	class StreamBuf : public IStreamBuf {
		std::streambuf *fwd;
	public:
		StreamBuf(std::streambuf* forwardTo):fwd(forwardTo) { }
		size_t Write(const char *buf, size_t bytes) noexcept { return (size_t)fwd->sputn(buf, bytes); }
		size_t Read(char *buf, size_t bytes) noexcept { return (size_t)fwd->sgetn(buf, bytes); }
		size_t Seek(int offset, int dir, int mode) noexcept { return (size_t)fwd->pubseekoff((std::streamoff)offset, (std::ios_base::seekdir)dir, (std::ios_base::openmode)mode); }
		int Sync() noexcept { return fwd->pubsync(); }
	};

	using String = Value<IStr>;

	class IType;

	static inline std::string MoveString(String b) { return std::string(b->data(), b->data() + b->size()); }

	class Type : public Shared<const IType> {
	public:
		Type(const IType& it) :Shared<const IType>(&it) {}
		Type(const IType *i):Shared<const IType>(i) { }
		Type(const std::string& string) : Type(_GetString(string.c_str())) { }
		Type(const char* string) : Type(_GetString(string)) { }
		Type(long double val) : Type(_GetConstant(val)) {}

		inline operator const K3::Type&() const { return Get()->GetPimpl(); }
		inline std::string AsString() const { return MoveString(_CheckResult(Get()->_AsString())); }

		inline Type GetUserTypeContent() const { return _CheckResult(Get()->_GetUserTypeContent()); }
		inline Type GetUserTypeTag() const { return _CheckResult(Get()->_GetUserTypeTag()); }
		inline Type GetFirst() const { return _CheckResult(Get()->_GetFirst()); }
		inline Type GetRest() const { return _CheckResult(Get()->_GetRest()); }

		inline bool operator<(const Type& t) const { return Get()->operator<((const IType&)t); }
		inline bool operator==(const Type& t) const { return Get()->operator==((const IType&)t); }
		inline bool operator!=(const Type& t) { return !operator==(t); }

		inline void ToStream(std::streambuf* buffer, const void *data, int printMode = IType::KronosWithParens) const {
			StreamBuf buf(buffer); Get()->ToStream(buf, data, printMode); _CheckLastError();
		}

		inline void ToStream(std::ostream& stream, const void *data, int printMode = IType::KronosWithParens) const {
			ToStream(stream.rdbuf(), data, printMode); _CheckLastError();
		}

		inline void ToStream(IStreamBuf& sb, const void *dataInstance, int printMode) const { Get()->ToStream(sb, dataInstance, printMode); _CheckLastError(); }

		inline bool IsUserType() const { return Get()->IsUserType(); }
		inline bool IsPair() const { return Get()->IsPair(); }
		inline bool IsTrue() const { return Get()->IsTrue(); }
		inline bool IsNil() const { return Get()->IsNil(); }
		inline bool IsConstant() const { return Get()->IsConstant(); }
		inline bool IsString() const { return Get()->IsString(); }
		inline bool IsFloat32() const { return Get()->IsFloat32(); }
		inline bool IsFloat64() const { return Get()->IsFloat64(); }
		inline bool IsInt32() const { return Get()->IsInt32(); }
		inline bool IsInt64() const { return Get()->IsInt64(); }

		inline long double AsConstant() const { return Get()->AsConstant(); }
		inline bool AsBool() const { return Get()->AsBool(); }

		inline size_t SizeOf() const { return Get()->SizeOf(); }
		inline size_t GetHash() const { return Get()->GetHash(); }
	};

	static Type GetString(const char* s) { return _CheckResult(_GetString(s)); }
	static Type GetConstant(long double c) { return _CheckResult(_GetConstant(c)); }
	static Type GetTuple(const IType& element, size_t count) { return _CheckResult(_GetTuple(element, count)); }
	static Type GetList(const IType& element, size_t count) { return _CheckResult(_GetList(element, count)); }
	static Type GetPair(const IType& fst, const IType& rst) { return _CheckResult(_GetPair(fst, rst)); }
	static Type Tuple(const Type& fst) { return fst; }
	template <typename... ARGS> static Type Tuple(const Type& fst, const Type& snd, const ARGS&... rst) { return _CheckResult(_GetPair(fst, Tuple(snd, rst...))); }
	static Type GetString(const std::string& s) { return _CheckResult(_GetString(s.c_str())); }
	static Type GetNil() { return _CheckResult(_GetNil()->Clone()); }
	static Type GetTrue() { return _CheckResult(_GetTrue()->Clone()); }
	static Type GetFloat32Ty() { return _CheckResult(_GetFloat32Ty()->Clone()); }
	static Type GetFloat64Ty() { return _CheckResult(_GetFloat64Ty()->Clone()); }
	static Type GetInt32Ty() { return _CheckResult(_GetInt32Ty()->Clone()); }
	static Type GetInt64Ty() { return _CheckResult(_GetInt64Ty()->Clone()); }

	/**
	* \brief Class represents compiled code, and can construct stateful instances
	*
	*/

	class StaticClass : Shared<IClass> {
	protected:
		using Shared<IClass>::Get;
	public:
		StaticClass(IClass *ic):Shared<IClass>(ic) { }

		inline Type GetListOfVars() const { return _CheckResult(Get()->_GetListOfVars()); }
		inline Type GetListOfTriggers() const { return _CheckResult(Get()->_GetListOfTriggers()); }
		inline Type TypeOfResult() const { return _CheckResult(Get()->_TypeOfResult()); }
		inline Type ReactivityOfResult() const { return _CheckResult(Get()->_ReactivityOfResult()); }
		inline Type GetUndefinedVars() const { return _CheckResult(Get()->_GetUndefinedVars()); }
		inline Type TypeOfVar(const Type& key) const { return _CheckResult(Get()->_TypeOfVar(*key)); }
		inline size_t GetVarOffset(const Type& key) const { return _CheckResult(Get( )->_GetVarOffset(key)); }
		inline Type GetVarClock(const Type& key) const { return _CheckResult(Get()->_GetVarClock(key)); }
		inline TriggerCallback GetTrigger(const Type& key) const { return _CheckResult(Get()->_GetTrigger(key)); }

		inline std::string GetTriggerName(const Type& key) const { return MoveString(_CheckResult(Get()->_GetTriggerName(*key))); }

		inline bool HasVar(const Type& key) const { return _CheckResult(Get()->_HasVar(*key)); }
		inline void SetVar(const Type& key, const void *data) { Get()->_SetVar(*key, data); _CheckLastError(); }

		inline void Provide(const Type& key, const void *data) { if (HasVar(key)) SetVar(key, data); }

		inline void MakeStatic(const char *namespacePrefix,
			const char *fileType,
			std::ostream& writeToStream,
			BuildFlags flags = Default,
			const char *mcpu = "",
			const char *march = "",
			const char *mfeat = "") {
			StreamBuf tmp(writeToStream.rdbuf()); Get()->_MakeStatic(namespacePrefix, fileType, tmp, flags, mcpu, march, mfeat); _CheckLastError();
		}

		inline size_t SizeOfInstance() const { return _CheckResult(Get()->SizeOfInstance()); }

		inline size_t GetNumDependentSymbols() const { return _CheckResult(Get()->_GetNumDependentSymbols()); }
		inline std::string GetDependentSymbol(size_t i) const { auto str(Get()->_GetDependentSymbol(i)); _CheckLastError(); return str; }

		using Shared<IClass>::GetHash;
		using Shared<IClass>::operator bool;
		bool operator==(const StaticClass& rhs) const { return Get()->_Equals(rhs.Get()); }
	};

	/**
	* \brief Represents a clock source in an instance of a Kronos DSP
	*
	*/
	class IInstance;
	struct Trigger {
		Trigger(IInstance* i = nullptr, TriggerCallback cb = nullptr) :instance(i), callback(cb) {}
		IInstance *instance;
		TriggerCallback callback;
		inline operator bool() const { return callback && instance; }
		inline void Go(void *outputBuffer, int32_t loopCount) noexcept { callback(instance->_GetStateBlob(), outputBuffer, loopCount); }
		inline bool operator==(const Trigger& rhs) const { return callback == rhs.callback && instance == rhs.instance; }
		inline bool operator!=(const Trigger& rhs) const { return !operator==(rhs); }
		inline Type GetOutputType() const { return StaticClass(instance->_TypeOf()).TypeOfResult(); }
	};

	/**
	* \brief Represents a live variable in an instance of a Kronos DSP
	*
	*/
	struct Var {
		IInstance *instance;
		Type key;
		size_t offset;
		Var(IInstance* i = nullptr, const Type& k = GetNil(), size_t idx = 0) :instance(i), key(k), offset(i?i->_TypeOf()->_GetVarOffset(k):-1) {}
		operator bool() const { return instance && key.IsNil() == false; }
		inline Type GetType() const { return StaticClass(instance->_TypeOf()).TypeOfVar(key); }
		inline Type GetStreamClock() const { return StaticClass(instance->_TypeOf()).GetVarClock(key); }
		inline const void *GetData() { return *GetSlotPointer(instance, offset); }
        inline const void* const GetData() const { return *GetSlotPointer(instance, offset); }
		inline void SetData(const void *data) { *GetSlotPointer(instance, offset) = data; }
		static inline void const**const GetSlotPointer(void *blob, size_t index) { return (void const **)((char *)blob + index); }
		static inline void const**const GetSlotPointer(IInstance* instance, size_t index) { return GetSlotPointer(instance->_GetStateBlob(), index); }
		inline bool operator==(const Var& rhs) const { return instance == rhs.instance && key == rhs.key; }
		inline bool operator!=(const Var& rhs) const { return !operator==(rhs); }

		/// convenience typed assignment
		template <typename T> Var& operator=(const T& data) {
			assert(GetData() && sizeof(T) >= GetType()->SizeOf() && "Buffer underflow");
			SetData((void*)&data);
			return *this;
		}

		/// convenience typed accessor
		template <typename T> const T& GetAs() const {
			assert(GetData() && sizeof(T) <= GetType()->SizeOf() && "Buffer overflow");
			return *(T*)GetData();
		};
	};

	class Input {
		Var data;
		Trigger callback;
	public:
		Input() {}
		Input(Var data, Trigger trigger) :data(data), callback(trigger) {}
		template <typename OUT> void Process(OUT out, const void* from) {
			if (data) {
				data.SetData((void*)from);
			}
			if (callback) {
				Tick(out);
			}
		}

		void Process(std::vector<char>& outBuf, const void* from) {
			outBuf.resize(callback.GetOutputType().SizeOf());
			Process(outBuf.data(), from);
		}

		template <typename T> void operator=(const T* from) {
			Process(nullptr, (const void*)from);
		}

		operator bool() const { return data && callback; }
		Type TypeOf() const { return data.GetType(); }

		void Tick(std::nullptr_t) {
			void *scratch = alloca(callback.GetOutputType().SizeOf());
			callback.Go(scratch, 1);
		}

		void Tick(void *output) {
			callback.Go(output, 1);
		}

		template <typename T> operator T() const { return data.GetAs<T>(); }
	};

	/**
	* \brief Instance represents a live instance of a compiled Kronos class
	*
	*/

	class Instance : Shared<IInstance> {
	public:
		Instance(IInstance *i = nullptr):Shared<IInstance>(i) { }
		inline Trigger GetTrigger(const Type& ID) { return Trigger( Get(), TypeOf().GetTrigger(ID) );  }
		inline Var GetVar(const Type& ID) { return TypeOf().HasVar(ID) ? Var( Get(), ID, TypeOf().GetVarOffset(ID) ) : Var();  }

		inline void Reset(const void *withArgument = nullptr) { Get()->Reset(withArgument); }
		inline void Evaluate(const void *input, void *output) { Get()->Evaluate(input, output); }

		inline Input operator[](const Type& ID) { return Input(GetVar(ID),GetTrigger(ID)); }
		inline Input operator[](const char *ID) { return operator[](GetString(ID)); }
		inline Input operator[](const std::string& ID) { return operator[](GetString(ID)); }
		inline StaticClass TypeOf() const { return Get()->_TypeOf(); }
		inline Instance Clone(bool withCurrentState, void* constructInPlace = nullptr) { return Instance(_CheckResult(Get()->_Clone(withCurrentState, constructInPlace))); }
		inline void CopyStateFrom(Instance& src) { Get()->_SetStateBlob(src.Get()->_GetStateBlob()); }
		inline void SetStateBlob(const void *from) { Get()->_SetStateBlob(from); }
		inline void* GetStateBlob() { return Get()->_GetStateBlob(); }

		bool operator==(const Instance& rhs) const { return Get( ) == rhs.Get( ); }
		bool operator!=(const Instance& rhs) const { return Get( ) != rhs.Get( ); }
		operator bool( ) const { return Get( ) != nullptr; }
		using Shared<IInstance>::GetHash;
	};

	class Class : public StaticClass {
	public:
		Class(IClass* ic = nullptr):StaticClass(ic) { }
		inline Instance Cons(const void *withArg = nullptr) { return _CheckResult(Get()->_Cons(withArg)); }
		inline Instance Cons(void *withMemory, const void *withArg) { return _CheckResult(Get()->_Cons(withMemory, withArg)); }
	};

	template <typename ITER> class seq_view {
		ITER b, e;
	public:
		seq_view(ITER b, ITER e):b(b), e(e) { }
		ITER begin() const { return b; }
		ITER end() const { return e; }
		size_t size() const { return e - b; }
	};
    
    using GenericGraph = Shared<const IGenericGraph>;

    class TypedGraph : protected Shared<const ITypedGraph> {
    public:
        TypedGraph(const ITypedGraph* ptr):Shared(ptr) {}
        Type TypeOfArgument() const { return Type(Get()->_TypeOfArgument()); }
        Type TypeOfResult() const { return Type(Get()->_TypeOfResult()); }
        const ITypedGraph* Get() const { return Shared::Get(); }
    };

	class StringSet : Shared<const IStringSet> {
		using LoopBody = std::function<void(const char*)>;
		static void KRONOS_ABI_FN ForeachForward(void* user, const char *str) {
			auto fn = (LoopBody*)user; (*fn)(str);
		}
	public:
		StringSet(const IStringSet* ss = nullptr) :Shared(ss) {}
		inline int Size() const { return _CheckResult(Get()->_Size()); }
		inline void ForEach(LoopBody body) const { Get()->_ForEach(ForeachForward, &body); _CheckLastError(); }
		inline bool Contains(const char *str) const { return _CheckResult(Get()->_Contains(str)); }
		inline bool Intersects(const StringSet& other) const { return _CheckResult(Get()->_Intersects(other.Get())); }
		operator bool() const { return Get() != nullptr; }
	};
    
    class Context : Shared<IContext> {
        using ImmediateHandler = std::function<void(const char*, GenericGraph)>;
        static void KRONOS_ABI_FN ImmediateHandlerForwarder(void* userdata, const char* sym, const IGenericGraph* ig) {
            auto imptr = (ImmediateHandler*)userdata; (*imptr)(sym,ig);
        }
	public:
		inline StringSet DrainRecentSymbolChanges() { return _CheckResult(Get()->_DrainRecentSymbolChanges()); }
		inline StringSet GetResolutionTrace() const { return _CheckResult(Get()->_GetResolutionTrace()); }

        inline void Parse(const char *source, bool REPLMode, ImmediateHandler h) {
            Get()->_Parse(source, REPLMode, ImmediateHandlerForwarder, &h);
            _CheckLastError();
        }
        
        inline GenericGraph Parse(const char *source, bool REPLMode = false) {
            GenericGraph gg;
            Parse(source,REPLMode,[&gg](const char*,GenericGraph g) mutable { gg = g; });
            if (!gg) throw RuntimeError("Expression was empty.");
            return gg;
        }
        
        inline TypedGraph Specialize(GenericGraph g, const Type& argumentType, std::ostream* log, int logLevel) {
            if (log) {
                StreamBuf tmp(log->rdbuf());
                return _CheckResult(Get()->_Specialize(g.Get(), argumentType, &tmp, logLevel));
            } else {
                return _CheckResult(Get()->_Specialize(g.Get(), argumentType, nullptr, logLevel));
            }
        }
        
        inline Class Make(const char *engine, TypedGraph g, BuildFlags flags = Default, const char* triple = "host",
                          const char* mcpu = "host",
                          const char* features = "") {
            return _CheckResult(Get()->_Make(engine, g.Get(), triple, mcpu, features, flags));
        }
        
        inline Class Make(const char* engine, const char* source, const Type& argumentType, std::ostream* log, int logLevel, BuildFlags flags = Default, const char* triple = "host",
                          const char* mcpu = "host",
                          const char* features = "") {
            return Make(engine, Specialize(Parse(source), argumentType, log, logLevel), flags, triple, mcpu, features);
        }

		inline Type DeriveType(const char *source, const Type& argumentType, std::ostream* log, int logLevel) {
            return Specialize(Parse(source), argumentType, log, logLevel).TypeOfResult();
		}
        
		using Shared<IContext>::operator bool;
		Context(IContext *ic = nullptr) : Shared<IContext>(ic) { }
		inline std::string GetModuleAndLineNumberText(const char* offset) { return MoveString(_CheckResult(Get()->_GetModuleAndLineNumberText(offset))); }
		inline std::string ShowModuleLine(const char* offset) { return MoveString(_CheckResult(Get()->_ShowModuleLine(offset))); }

		static const char *view_str(const char *s) { return s; }
		static const char *view_str(const std::string& s) { return s.c_str(); }

		template <typename CONTAINER> void SetModuleImportPaths(const CONTAINER& c) {
			std::vector<const char*> paths; for (auto& p : c) paths.push_back(view_str(p));
			Get()->_SetModuleImportPaths(paths.data(), paths.data() + paths.size());
		}

		template <typename CONTAINER> void AddModuleImportPaths(const std::string& path) {
			std::vector<const char*> paths{ path.c_str() };
			GetModuleImportPaths(paths);
			Get()->_SetModuleImportPaths(paths.data(), paths.data() + paths.size());
		}

		template <typename CONTAINER> void GetModuleImportPaths(CONTAINER& c) {
			const char *const*b(nullptr), *const*e(nullptr);
			Get()->_GetModuleImportPaths(b, e);
			for (auto i = b;i != e;++i) c.push_back(*i);
		}

		inline void ImportFile(const std::string& modulePath, bool search = false) { Get()->_ImportFile(modulePath.c_str(), search); _CheckLastError(); }
		inline void ImportBuffer(const std::string& module, const std::string& sourceCode, const std::string& inPackage = "", bool allowRedefinition = false) { Get( )->_ImportBuffer(module.c_str( ), sourceCode.c_str( ), inPackage.c_str( ), allowRedefinition); _CheckLastError(); }
	};

	static std::string GetUserPath() { return MoveString(_GetUserPath()); }
	static std::string GetSharedPath() { return MoveString(_GetSharedPath()); }

	// propagate exceptions to the client ABI
	static inline void _CheckLastError() {
		IError *err = _GetAndResetThreadError();
		if (err == nullptr) return;
		std::string msg = err->GetErrorMessage(); auto code = err->GetErrorCode();
		auto pos = err->GetSourceFilePosition(); auto ty = err->GetErrorType();
		err->Delete();
		switch (ty) {
		case IError::InternalError: throw InternalError(msg.c_str());
		case IError::RuntimeError: throw RuntimeError(msg.c_str());
		case IError::SyntaxError: throw SyntaxError(code, msg.c_str(), pos);
		case IError::TypeError: throw TypeError(code, msg.c_str(), pos);
		default: assert(0 && "ICE: Unexpected error of unknown type");
		}
	}


	static Context CreateContext() {
		return _CheckResult(_CreateContext());
	}

	class TypeIter {
		Type t;
	public:
		TypeIter(Type t):t(std::move(t)) { }
		TypeIter operator++() { t = t.GetRest(); return *this; }
		TypeIter operator++(int) { auto tmp(*this); operator++(); return tmp; }
		Type operator*() const { return t.GetFirst(); }
		bool operator==(const TypeIter& r) { return *t == *r.t; }
		bool operator!=(const TypeIter& r) { return !operator==(r); }
	};

	static Kronos::TypeIter end(const Kronos::Type) { return Kronos::GetNil(); }
	static Kronos::TypeIter begin(const Kronos::Type &t) { return t.IsPair() ? t : end(t); }
}

namespace std {
	template <> struct hash<Kronos::Type> { size_t operator()(const Kronos::Type& t) const { return t->GetHash(); } };
	template <> struct hash<Kronos::StaticClass> { size_t operator()(const Kronos::StaticClass& ks) const { return ks.GetHash(); } };
	template <> struct hash<Kronos::Instance> { size_t operator()(const Kronos::Instance& i) const { return i.GetHash(); } };
	static std::ostream& operator<<(std::ostream& s, const Kronos::Type& t) { t.ToStream(s,nullptr,false); return s; }
	template <> struct iterator_traits<Kronos::TypeIter> {
		using value_type = Kronos::Type;
		using reference = Kronos::Type&;
		using pointer = Kronos::Type*;
		using iterator_category = forward_iterator_tag;
	};
}


