#pragma once
#ifndef __KRONOS_ABI_H
#define __KRONOS_ABI_H

/*
 *	Kronos ABI
 *
 *  This header defines the data types that may cross the shared library boundary.
 *  ABI compatibility level similar to that required by COM is required between
 *  the client and the library
 *
 */

#include <cstdint>
#include <cstddef> 

#ifdef WIN32
    #ifdef _WIN64
        #define KRONOS_ABI_FN
        #define KRONOS_ABI_MEMFN
    #else
        #define KRONOS_ABI_FN __fastcall
        #define KRONOS_ABI_MEMFN __thiscall
    #endif
    #ifdef K3_EXPORTS
        #define KRONOS_ABI extern "C" __declspec(dllexport)
    #elif K3_IMPORTS
        #define KRONOS_ABI extern "C" __declspec(dllimport)
    #else
        #define KRONOS_ABI extern "C"
    #endif
#else
    #define KRONOS_ABI_FN
    #define KRONOS_ABI_MEMFN
    #ifdef K3_EXPORTS
        #define KRONOS_ABI extern "C" __attribute__ ((visibility("default"),cdecl))
    #else
        #define KRONOS_ABI extern "C" __attribute__ ((cdecl))
    #endif
#endif

#define KRONOS_INT std::int32_t

#define ABI KRONOS_ABI
#define MEMBER KRONOS_ABI_MEMFN
#define FUNCTION KRONOS_ABI_FN
#define INT KRONOS_INT

namespace K3 {
	class Type;
    namespace Nodes {
        class Generic;
        class Typed;
    }
}

namespace CmdLine {
	class IRegistry;
};

namespace Kronos {
	class IShared {
		virtual void MEMBER Attach() const noexcept = 0;
		virtual void MEMBER Detach() const noexcept = 0;
	protected:
		virtual MEMBER ~IShared() { }
	public:
		virtual void MEMBER Delete() const noexcept = 0;
		inline void Retain() const { Attach(); }
		inline void Release() const { Detach(); }
	};

	class IValue {
	protected:
		virtual MEMBER ~IValue() { }
	public:
		virtual void MEMBER Delete() noexcept = 0;
		virtual IValue* MEMBER Clone() const noexcept = 0;
	};

	class IStr : public IValue {
	public:
		virtual const char* MEMBER c_str() const noexcept = 0;
		virtual const char* MEMBER data() const noexcept = 0;
		virtual size_t MEMBER size() const noexcept = 0;
	};

	class IStreamBuf {
	public:
		virtual ~IStreamBuf() { }
		virtual size_t MEMBER Seek(INT, INT, INT) noexcept = 0;
		virtual size_t MEMBER Write(const char* buffer, size_t numBytes) noexcept = 0;
		virtual size_t MEMBER Read(char* buffer, size_t numBytes) noexcept = 0;
		virtual INT MEMBER Sync() noexcept = 0;
	};

	class IError : public IValue {
	public:
		enum ErrorType {
			InternalError,
			RuntimeError,
			SyntaxError,
			TypeError
		};
		virtual const char* MEMBER GetErrorMessage() const noexcept = 0;
		virtual ErrorType MEMBER GetErrorType() const noexcept = 0;
		virtual const char* MEMBER GetSourceFilePosition() const noexcept = 0;
		virtual INT MEMBER GetErrorCode() const noexcept = 0;
		virtual void MEMBER Delete() noexcept = 0;
		virtual IError* MEMBER Clone() const noexcept = 0;
	};

	class IInternalError : public IError { 
	public:
		static ErrorType GetClassErrorType() { return InternalError; }
	};

	class IRuntimeError : public IError { 
	public:
		static ErrorType GetClassErrorType() { return RuntimeError; }
	};

	class IProgramError : public IError { };

	class ISyntaxError : public IProgramError { 
	public:
		static ErrorType GetClassErrorType() { return SyntaxError; }
	};
	class ITypeError : public IProgramError { 
	public:
		static ErrorType GetClassErrorType() { return TypeError; }
	};

	class IType : public IShared {
	public:
		virtual IType* Clone() const noexcept = 0;
		virtual IStr* MEMBER _AsString() const noexcept = 0;
		virtual void MEMBER ToStream(IStreamBuf&, const void *dataInstance, INT printMode) const noexcept = 0;
		virtual IType* MEMBER _GetUserTypeContent() const noexcept = 0;
		virtual IType* MEMBER _GetUserTypeTag() const noexcept = 0;
		virtual IType* MEMBER _GetFirst() const noexcept = 0;
		virtual IType* MEMBER _GetRest() const noexcept = 0;

		virtual const K3::Type& MEMBER GetPimpl() const noexcept = 0;

		virtual bool MEMBER IsUserType() const noexcept = 0;
		virtual bool MEMBER IsPair() const noexcept = 0;
		virtual bool MEMBER IsTrue() const noexcept = 0;
		virtual bool MEMBER IsNil() const noexcept = 0;
		virtual bool MEMBER IsConstant() const noexcept = 0;
		virtual bool MEMBER IsString() const noexcept = 0;
		virtual bool MEMBER IsFloat32() const noexcept = 0;
		virtual bool MEMBER IsFloat64() const noexcept = 0;
		virtual bool MEMBER IsInt32() const noexcept = 0;
		virtual bool MEMBER IsInt64() const noexcept = 0;

		virtual long double MEMBER AsConstant() const noexcept = 0;
		virtual bool MEMBER AsBool() const noexcept = 0;

		virtual size_t MEMBER SizeOf() const noexcept = 0;
		virtual size_t MEMBER GetHash() const noexcept = 0;
		virtual IType& MEMBER operator=(const IType&) noexcept = 0;

		virtual bool MEMBER operator<(const IType&) const noexcept = 0;
		virtual bool MEMBER operator==(const IType&) const noexcept = 0;

		enum PrintMode {
			KronosWithParens,
			KronosWithoutParens,
			XML,
			EDN
		};
	};

	class IUserException : public IError {
	public:
		virtual const IType& MEMBER _GetUserException() const noexcept = 0;
	};


	using TriggerCallback = void(*)(void *instance, void *outputBuffer, std::int32_t loopCount);

	class IClass;

	class IInstance : public IShared {
	public:
		virtual IClass* MEMBER _TypeOf() const noexcept = 0;
		virtual void MEMBER Reset(const void *withArgument) noexcept = 0;
		virtual void MEMBER Evaluate(const void *input, void *output) noexcept = 0;
		virtual IInstance* MEMBER _Clone(bool state, void *mem) noexcept = 0;
		virtual void MEMBER _SetStateBlob(const void* from) noexcept = 0;
		virtual void* MEMBER _GetStateBlob() noexcept = 0;
	};

	enum BuildFlags {
		Default = 0,
		OmitEvaluate = 1,
		StrictFloatingPoint = 2,
		EmulateFloatingPoint = 4,
		OmitReactiveDrivers = 8
	};

	class IClass : public IShared {
	public:
		virtual IInstance* MEMBER _Cons(const void *withArg) const noexcept = 0;
		virtual IInstance* MEMBER _Cons(void *withMemory, const void* withArg) const noexcept = 0;

		virtual IType* MEMBER _GetListOfVars() const noexcept = 0;
		virtual IType* MEMBER _GetListOfTriggers() const noexcept = 0;
		virtual TriggerCallback MEMBER _GetTrigger(const IType& key) const noexcept = 0;
		virtual size_t MEMBER _GetVarOffset(const IType& key) const noexcept = 0;
		virtual IType* MEMBER _GetVarClock(const IType& key) const noexcept = 0;
		virtual IType* MEMBER _TypeOfResult() const  noexcept = 0;
		virtual IType* MEMBER _ReactivityOfResult() const noexcept = 0;
		virtual IType* MEMBER _GetUndefinedVars() const noexcept = 0;
		virtual IType* MEMBER _TypeOfVar(const IType& key) const noexcept = 0;

		virtual bool MEMBER _HasVar(const IType& key) const noexcept = 0;
		virtual void MEMBER _SetVar(const IType& key, const void* data) noexcept = 0;

		virtual IStr* MEMBER _GetTriggerName(const IType& key) const noexcept = 0;

		virtual void MEMBER _MakeStatic(const char *namespacePrefix,
			const char *fileType,
			IStreamBuf& writeToStream,
			BuildFlags flags,
			const char *mcpu,
			const char *march,
			const char *mfeatures)  noexcept = 0;

		virtual size_t MEMBER SizeOfInstance() const noexcept = 0;
		virtual size_t MEMBER _GetNumDependentSymbols() const noexcept = 0;
		virtual const char * MEMBER _GetDependentSymbol(size_t) const noexcept = 0;
		virtual bool MEMBER _Equals(const IClass*) const noexcept = 0;
	};
    
    class IGenericGraph : public IShared {
    public:
        virtual const K3::Nodes::Generic* MEMBER Get() const noexcept = 0;
    };
    
    class ITypedGraph : public IShared {
    public:
        virtual const K3::Nodes::Typed* MEMBER Get() const noexcept = 0;
        virtual IType* MEMBER _TypeOfResult() const noexcept = 0;
        virtual IType* MEMBER _TypeOfArgument() const noexcept = 0;
        virtual const K3::Type* MEMBER _InternalTypeOfResult() const noexcept = 0;
        virtual const K3::Type* MEMBER _InternalTypeOfArgument() const noexcept = 0;
    };
    
    using ImmediateExpressionHandler = void FUNCTION (void*,const char*,const IGenericGraph*);
 
	using StringSetIterator = void FUNCTION (void*, const char*);

	class IStringSet : public IShared {
	public:
		virtual INT MEMBER _Size() const noexcept = 0;
		virtual void MEMBER _ForEach(StringSetIterator, void *userdata) const noexcept = 0;
		virtual bool MEMBER _Contains(const char *str) const noexcept = 0;
		virtual bool MEMBER _Intersects(const IStringSet* with) const noexcept = 0;
	};
	
	class ICompilerInterface : public IShared {
	public:
        virtual void MEMBER _Parse(const char *source, bool REPLMode, ImmediateExpressionHandler, void *userdata) noexcept = 0;
        virtual const ITypedGraph* MEMBER _Specialize(const IGenericGraph*, const IType& argumentType, IStreamBuf* log, INT logLevel) noexcept = 0;
		virtual IClass* MEMBER _Make(
			const char* engine,
            const ITypedGraph*,
            const char* triple,
			const char* mcpu,
			const char* features,
			BuildFlags flags)  noexcept = 0;
		virtual const IStringSet* MEMBER _GetResolutionTrace() const noexcept = 0;
		virtual const IStringSet* MEMBER _DrainRecentSymbolChanges() noexcept = 0;
	};

	class IContext : public ICompilerInterface {
	public:
		virtual void MEMBER _GetModuleImportPaths(const char*const*& first, const char*const*& end) noexcept = 0;
		virtual void MEMBER _SetModuleImportPaths(const char*const* first, const char*const* end) noexcept = 0;
		virtual void MEMBER _ImportFile(const char *modulePath, bool search)  noexcept = 0;
		virtual void MEMBER _ImportBuffer(const char* module, const char* sourceCode, const char* inPackage, bool allowRedefinition)  noexcept = 0;
		virtual IStr* MEMBER _GetModuleAndLineNumberText(const char*)  noexcept = 0;
		virtual IStr* MEMBER _ShowModuleLine(const char*) noexcept = 0;
	};

	ABI const char* FUNCTION GetVersionString( ) noexcept;
	ABI void FUNCTION AddBackendCmdLineOpts(CmdLine::IRegistry& MasterRegistry) noexcept;

	ABI IStr* FUNCTION _GetUserPath()  noexcept;
	ABI IStr* FUNCTION _GetSharedPath() noexcept;
	ABI IContext* FUNCTION _CreateContext() noexcept;

	ABI IType* FUNCTION _GetString(const char*) noexcept;
	ABI IType* FUNCTION _GetConstant(long double) noexcept;
	ABI IType* FUNCTION _GetTuple(const IType& element, size_t count) noexcept;
	ABI IType* FUNCTION _GetList(const IType& element, size_t count) noexcept;
	ABI IType* FUNCTION _GetPair(const IType& fst, const IType& rst) noexcept;
	ABI const IType* FUNCTION _GetNil() noexcept;
	ABI const IType* FUNCTION _GetTrue() noexcept;
	ABI const IType* FUNCTION _GetFloat32Ty() noexcept;
	ABI const IType* FUNCTION _GetFloat64Ty() noexcept;
	ABI const IType* FUNCTION _GetInt32Ty() noexcept;
	ABI const IType* FUNCTION _GetInt64Ty() noexcept;
	ABI IType* FUNCTION _ConvertToABIType(const void *raw) noexcept;

	ABI IError* FUNCTION _GetAndResetThreadError() noexcept;
}
#undef ABI
#undef FUNCTION
#undef MEMBER
#undef INT
#endif
